/*****************************************************************************

       Copyright   1995, 1996 Digital Equipment Corporation,
                       Maynard, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, provided  
that the copyright notice and this permission notice appear in all copies  
of software and supporting documentation, and that the name of Digital not  
be used in advertising or publicity pertaining to distribution of the software 
without specific, written prior permission. Digital grants this permission 
provided that you prominently mark, as not part of the original, any 
modifications made to this software or documentation.

Digital Equipment Corporation disclaims all warranties and/or guarantees  
with regard to this software, including all implied warranties of fitness for 
a particular purpose and merchantability, and makes no representations 
regarding the use of, or the results of the use of, the software and 
documentation in terms of correctness, accuracy, reliability, currentness or
otherwise; and you rely on the software, documentation and results solely at 
your own risk. 

******************************************************************************/

/*****************************************************************************
*
* Copyright 1996, 1997, 1998 Hans Reiser, see /usr/src/linux/reiserfs/README
*                     for licensing and copyright details
*
*****************************************************************************/


/* This is a set of functions that provides minimal filesystem
 * functionality to the Linux bootstrapper.  All we can do is
 * open and read files... but that's all we need 8-)
 */

/*>>>>>>>>>>>>*/
/*#include <linux/module.h>*/
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <linux/reiserfs_fs.h>
#include <linux/reiserfs_fs_i.h>
#include <linux/reiserfs_fs_sb.h>
#include <linux/locks.h>
#include <linux/init.h>


#define KEY_ARRAY_SIZE 100

static struct {
    unsigned long objectid;
    unsigned long dirid;
    kdev_t dev ;
} g_key_array [KEY_ARRAY_SIZE] = {{0,},};


/*<<<<<<<<<<<<*/


#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/kdev_t.h>

#include "milo.h"
#include "fs.h"
#include "ext2.h"


#define REISERFS_BLOCK_SIZE 4096
#define REISERFS_MAX_BLOCK_SIZE REISERFS_BLOCK_SIZE

#define MAX_OPEN_FILES		4

int fd = -1;
struct reiserfs_super_block sb;
/*struct reiserfs_group_desc *gds = NULL;*/
int ngroups = 0;
int blocksize;			/* Block size of this fs */
int directlim;			/* Maximum direct blkno */
int ind1lim;			/* Maximum single-indir blkno */
int ind2lim;			/* Maximum double-indir blkno */
int ptrs_per_blk;		/* ptrs/indirect block */
char filename[256];

char blkbuf[REISERFS_MAX_BLOCK_SIZE];
long cached_iblkno = -1;

char iblkbuf[REISERFS_MAX_BLOCK_SIZE];
long cached_diblkno = -1;

char diblkbuf[REISERFS_MAX_BLOCK_SIZE];

static struct inode_table_entry {
	struct reiserfs_inode inode;
	int inumber;
	int free;
	unsigned short old_mode;
} inode_table[MAX_OPEN_FILES];


/* Initialize an reiserfs filesystem; this is sort-of the same idea as
 * "mounting" it.  Read in the relevant control structures and 
 * make them available to the user.  Returns 0 if successful, -1 on
 * failure.
 */

/* forward declarations */
static int reiserfs_bread(int fd, long blkno, char *buffer);
static void reiserfs_breadi(struct reiserfs_inode *ip, long blkno, char *buffer);
static int reiserfs_mount(long device, int quiet);
static int reiserfs_open(char *filename);
static void reiserfs_ls(char *dirname, char *devname);
static void reiserfs_close(int fd);
static long reiserfs_fdsize(int fd);

struct bootfs reiserfs =
{
	"reiserfs",
	0,
	0,

	reiserfs_mount,
	reiserfs_open,
	reiserfs_fdsize,
	reiserfs_bread,
	reiserfs_ls,
	reiserfs_close,
};







/*>>>>>>>>>>>>>>>*/



void store_key (struct super_block *s, struct key * key)
{
    int i;

    for (i = 0; i < sizeof (g_key_array) / sizeof (g_key_array[0]); i ++) {
	if (g_key_array[i].objectid == 0) {
	    g_key_array[i].dirid = key->k_dir_id;
	    g_key_array[i].objectid = key->k_objectid;
	    g_key_array[i].dev = s->s_dev ;
	    return;
	}
    }
    reiserfs_warning ("vs-13042: store_key: table of keys is full\n");
}


#if 0
static int read_super_block (struct super_block * s, int size)
{
	struct buffer_head * bh;
	struct reiserfs_super_block * rs;
	int repeat ;
	
	repeat = 0 ;
	bh = bread (s->s_dev, (REISERFS_DISK_OFFSET_IN_BYTES / size), size);
  
	if (!bh)
		{
			printk("reiserfs_read_super: unable to read superblock on dev %s\n", kdevname(s->s_dev));
			return 1;
		}
	
	rs = (struct reiserfs_super_block *)bh->b_data;
	if (strncmp (rs->s_magic,  REISERFS_SUPER_MAGIC_STRING, strlen ( REISERFS_SUPER_MAGIC_STRING)))
		{
			brelse (bh);
			printk ("reiserfs_read_super: can't find a reiserfs filesystem on dev %s.\n", kdevname(s->s_dev));
			return 1;
		}

	s->s_blocksize = le16_to_cpu (rs->s_blocksize);
	s->s_blocksize_bits = 0;
	while ((1 << s->s_blocksize_bits) != s->s_blocksize)
		s->s_blocksize_bits ++;

	if (size != s->s_blocksize)
		{
			brelse (bh);
			set_blocksize (s->s_dev, s->s_blocksize);
			bh = reiserfs_bread (s->s_dev,  (REISERFS_DISK_OFFSET_IN_BYTES / s->s_blocksize), s->s_blocksize, &repeat);
			if (!bh)
				{
					printk("reiserfs_read_super: unable to read superblock on dev %s\n", kdevname(s->s_dev));
					return 1;
				}
			
			rs = (struct reiserfs_super_block *)bh->b_data;
			if (strncmp (rs->s_magic,  REISERFS_SUPER_MAGIC_STRING, strlen ( REISERFS_SUPER_MAGIC_STRING)) ||
			  le16_to_cpu (rs->s_blocksize) != s->s_blocksize)
				{
					brelse (bh);
					printk ("reiserfs_read_super: can't find a reiserfs filesystem on dev %s.\n", kdevname(s->s_dev));
					return 1;
				}
		}
	/* must check to be sure we haven't pulled an old format super out of the
	** old format's log.  This is a kludge of a check, but it will work.  
	** If block we've just read in is inside the journal for that
	** super, it can't be valid.
	*/
	if (bh->b_blocknr >= rs->s_journal_block && 
		bh->b_blocknr < (rs->s_journal_block + JOURNAL_BLOCK_COUNT))
		{
			brelse(bh) ;
			printk("super-459: reiserfs_read_super: super found at block %lu is within its own log.  It must not be of this format type.\n", bh->b_blocknr) ;
			return 1 ;
		}

	SB_BUFFER_WITH_SB (s) = bh;
	SB_DISK_SUPER_BLOCK (s) = rs;
	s->s_op = &reiserfs_sops;
	return 0;
}
#endif

struct key root_key = {REISERFS_ROOT_PARENT_OBJECTID, REISERFS_ROOT_OBJECTID, 0, 0};

static int read_bitmaps (struct super_block * s)
{
  int i, repeat, bmp, dl ;
  struct reiserfs_super_block * rs = SB_DISK_SUPER_BLOCK(s);

  repeat = 0 ;
  /* read true bitmap block */
  SB_AP_BITMAP (s) = reiserfs_kmalloc (sizeof (struct buffer_head *) * le16_to_cpu (rs->s_bmap_nr), GFP_KERNEL, s);

  if (SB_AP_BITMAP (s) == 0)
	  return 1;

  memset (SB_AP_BITMAP (s), 0, sizeof (struct buffer_head *) * le16_to_cpu (rs->s_bmap_nr));

                /* read bitmap blocks */
				/* reiserfs leaves the first 64k unused
				   so that any partition labeling scheme
				   currently used will have enough
				   space. Then we need one block for the
				   super.  -Hans */
  bmp = (REISERFS_DISK_OFFSET_IN_BYTES / s->s_blocksize) + 1;	/* first of bitmap blocks */
  SB_AP_BITMAP (s)[0] = reiserfs_bread (s->s_dev, bmp, s->s_blocksize, &repeat);

  if(!SB_AP_BITMAP(s)[0])
	  return 1;
  for (i = 1, bmp = dl = rs->s_blocksize * 8; i < le16_to_cpu (rs->s_bmap_nr); i ++)
	  {
		  SB_AP_BITMAP (s)[i] = reiserfs_bread (s->s_dev, bmp, s->s_blocksize, &repeat);
		  if (!SB_AP_BITMAP (s)[i])
			  return 1;
		  bmp += dl;
	  }

  return 0;
}

#if 0
static int read_old_bitmaps (struct super_block * s)
{
  int i, repeat ;
  struct reiserfs_super_block * rs = SB_DISK_SUPER_BLOCK(s);
  int bmp1 = (REISERFS_OLD_DISK_OFFSET_IN_BYTES / s->s_blocksize) + 1;  /* first of bitmap blocks */

  repeat = 0 ;
  /* read true bitmap */
  SB_AP_BITMAP (s) = reiserfs_kmalloc (sizeof (struct buffer_head *) * le16_to_cpu (rs->s_bmap_nr), GFP_KERNEL, s);
  if (SB_AP_BITMAP (s) == 0)
	  return 1;

  memset (SB_AP_BITMAP (s), 0, sizeof (struct buffer_head *) * le16_to_cpu (rs->s_bmap_nr));

  for (i = 0; i < le16_to_cpu (rs->s_bmap_nr); i ++)
	  {
		  SB_AP_BITMAP (s)[i] = reiserfs_bread (s->s_dev, bmp1 + i, s->s_blocksize, &repeat);
		  if (!SB_AP_BITMAP (s)[i])
			  return 1;
	  }
	
  return 0;
}

/* support old disk layout */
static int read_old_super_block (struct super_block * s, int size)
{
  struct buffer_head * bh;
  struct reiserfs_super_block * rs;
  int repeat ;

  printk("reiserfs_read_super: try to find super block in old location\n");
  repeat = 0 ;
  /* there are only 4k-sized blocks in v3.5.10 */
  if (size != REISERFS_OLD_BLOCKSIZE)
	  set_blocksize(s->s_dev, REISERFS_OLD_BLOCKSIZE);
  bh = bread (s->s_dev, 
  			  REISERFS_OLD_DISK_OFFSET_IN_BYTES / REISERFS_OLD_BLOCKSIZE, 
 		      REISERFS_OLD_BLOCKSIZE);
  if (!bh)
	  {
		  printk("reiserfs_read_super: unable to read superblock on dev %s\n", kdevname(s->s_dev));
		  return 1;
	  }

  rs = (struct reiserfs_super_block *)bh->b_data;
  if (strncmp (rs->s_magic,  REISERFS_SUPER_MAGIC_STRING, strlen ( REISERFS_SUPER_MAGIC_STRING)))
	  {
		  /* pre-journaling version check */
		  if(!strncmp((char*)rs + REISERFS_SUPER_MAGIC_STRING_OFFSET_NJ,
					  REISERFS_SUPER_MAGIC_STRING, strlen(REISERFS_SUPER_MAGIC_STRING)))
			  {
				  printk("reiserfs_read_super: a pre-journaling reiserfs filesystem isn't suitable there.\n");
				  brelse(bh);
				  return 1;
			  }
		  
		  brelse (bh);
		  printk ("reiserfs_read_super: can't find a reiserfs filesystem on dev %s.\n", kdevname(s->s_dev));
		  return 1;
	  }

  if(REISERFS_OLD_BLOCKSIZE != le16_to_cpu (rs->s_blocksize))
	  {
		  printk("reiserfs_read_super: blocksize mismatch, super block corrupted\n");
		  brelse(bh);
		  return 1;
	  }	

  s->s_blocksize = REISERFS_OLD_BLOCKSIZE;
  s->s_blocksize_bits = 0;
  while ((1 << s->s_blocksize_bits) != s->s_blocksize)
	  s->s_blocksize_bits ++;

  SB_BUFFER_WITH_SB (s) = bh;
  SB_DISK_SUPER_BLOCK (s) = rs;
  s->s_op = &reiserfs_sops;
  return 0;
}

#endif

struct super_block * reiserfs_read_super (struct super_block * s, void * data, int silent)
{
    int size;
    struct inode *root_inode;
    kdev_t dev = s->s_dev;
    int j;
    extern int *blksize_size[];
/*    struct reiserfs_transaction_handle th ;*/
    int old_format = 0;
    __u32 blocks;
    int jinit_done = 0 ;

    memset (&s->u.reiserfs_sb, 0, sizeof (struct reiserfs_sb_info));

#if 0
  if (parse_options ((char *) data, &(s->u.reiserfs_sb.s_mount_opt), &blocks) == 0)
	  {
		  s->s_dev = 0;
		  return NULL;
	  }
#endif

  MOD_INC_USE_COUNT;
  lock_super (s);

  if ( blksize_size[MAJOR(dev)] && blksize_size[MAJOR(dev)][MINOR(dev)] != 0)
	  {
		  /* as blocksize is set for partition we use it */
		  size = blksize_size[MAJOR(dev)][MINOR(dev)];
	  }
  else
	  {
		  size = BLOCK_SIZE;
		  set_blocksize (s->s_dev, BLOCK_SIZE);
	  }
  
    /* read block, containing reiserfs super block (it is stored at REISERFS_FIRST_BLOCK-th 1K block) */
  if ( read_super_block (s, size) )
	  {
		  if( read_old_super_block(s,size) )  goto error;
		  else                    	    old_format = 1;
	  }
  
/*    s->u.reiserfs_sb.s_mount_state = le16_to_cpu (SB_DISK_SUPER_BLOCK (s)->s_state); /* journal victim */
  s->u.reiserfs_sb.s_mount_state = REISERFS_VALID_FS ;


    /* reiserfs can not be mounted when it propably contains errors */


#if 0 /* journal victim */
  if (le16_to_cpu (SB_DISK_SUPER_BLOCK (s)->s_state) != REISERFS_VALID_FS)
	  {
		  printk ("reiserfs_read_super:  mounting unchecked fs, run reiserfsck first\n");
		  goto error;
	  }
#endif
  if (old_format ? read_old_bitmaps(s) : read_bitmaps(s))
	  { 
		  printk ("reiserfs_read_super: unable to read bitmap\n");
		  goto error;
	  }

  
#if 0
  if (journal_init(s))
	  {
		  printk("reiserfs_read_super: unable to initialize journal space\n") ;
			goto error ;
	  }
  else
	  {
		  jinit_done = 1 ; /* once this is set, journal_release must be called
							 ** if we error out of the mount 
							 */
	  }
  
  if (reread_meta_blocks(s))
	  {
		  printk("reiserfs_read_super: unable to reread meta blocks after journal init\n") ;
		  goto error ;
	  }
  
  if ( replay_only (s) ) 	goto error;
#endif
  
  /* VFS should do this for us, but we panic if the MS_RDONLY isn't
  ** set on a read only device, so I'm doing it again anyway. --clm
  */
  if (is_read_only(s->s_dev) && !(s->s_flags & MS_RDONLY))
	  {
		  printk("clm-4000: Detected readonly device, marking FS readonly\n") ;
		  s->s_flags |= MS_RDONLY ;
	  }
  /*s->s_op = &reiserfs_sops;*/
  
  /* get root directory inode */
  store_key (s, &root_key);
  root_inode = iget (s, root_key.k_objectid);
  forget_key (s, &root_key);
  if (!root_inode)
	  {
		  printk ("reiserfs_read_super: get root inode failed\n");
		  goto error;
	  }
  
  s->s_root = d_alloc_root(root_inode, NULL);  
  if (!s->s_root)
	  {
		  iput(root_inode);
		  goto error;
	  }
  
  // define and initialize hash function
  s->u.reiserfs_sb.s_hash_function = hash_function (s);
  if (s->u.reiserfs_sb.s_hash_function == NULL)
	  {
		  dput(s->s_root) ;
		  goto error ;
	  }
  
  
  if (!(s->s_flags & MS_RDONLY))
	  {
		  SB_DISK_SUPER_BLOCK (s)->s_state = cpu_to_le16 (REISERFS_ERROR_FS);
		  /* mark_buffer_dirty (SB_BUFFER_WITH_SB (s), 1); */
		  journal_begin(&th, s, 1) ;
		  journal_mark_dirty(&th, s, SB_BUFFER_WITH_SB (s));
		  journal_end(&th, s, 1) ;
		  s->s_dirt = 0;
	  }

  /*s->u.reiserfs_sb.unpreserve = dont_preserve (s) ? 0 : unpreserve;*/
  /* we have to do this to make journal writes work correctly */
  SB_BUFFER_WITH_SB(s)->b_end_io = reiserfs_end_buffer_io_sync ;
  
  unlock_super (s);
  print_credits() ;
    printk("%s\n", reiserfs_get_version_string()) ;
    return s;
	
  error:
	
#if 0
	if (jinit_done) /* kill the commit thread, free journal ram */
		{
			journal_release_error(NULL, s) ;
		}
#endif
	
	if (SB_DISK_SUPER_BLOCK (s))
		{
			for (j = 0; j < le16_to_cpu (SB_DISK_SUPER_BLOCK (s)->s_bmap_nr); j ++)
				{
					if ( SB_AP_BITMAP (s) )  brelse ( SB_AP_BITMAP (s)[j] );
				}
			if (SB_AP_BITMAP (s))
				reiserfs_kfree (SB_AP_BITMAP (s), sizeof (struct buffer_head *) * SB_BMAP_NR (s), s);
		}
    if (SB_BUFFER_WITH_SB (s))
		brelse(SB_BUFFER_WITH_SB (s));
    s->s_dev = 0;
    unlock_super(s);
    MOD_DEC_USE_COUNT;
	
    return NULL;
}

/*<<<<<<<<<<<<<<*/



static int reiserfs_mount(long device, int quiet)
{
	extern void set_blocksize(kdev_t dev, int size);
	
	struct buffer_head * bh;
	struct reiserfs_super_block * rs;
	int repeat ;
	
	repeat = 0 ;
	
	int i;
	long sb_offset;
	int sb_block = 16;
	
#ifdef DEBUG_REISERFS
	printk("reiserfs_mount(): called, device = 0x%04lx\n", device);
#endif
	/* Initialize the inode table */
	for (i = 0; i < MAX_OPEN_FILES; i++)
		{
			inode_table[i].free = 1;
			inode_table[i].inumber = 0;
		}

	/* set the device's block size */
	set_blocksize(device, REISERFS_BLOCK_SIZE);

	/* Read in the first superblock */
/*	sb_offset = sb_block *  REISERFS_BLOCK_SIZE;*/

	sb_offset = REISERFS_DISK_OFFSET_IN_BYTES

#ifdef DEBUG_REISERFS
	printk("Reading the first superblock() at offset 0x%lx\n",  sb_offset);
#endif

	size=REISERFS_BLOCK_SIZE;
/*
	if ( device_read (fd, (char *) bh, size, (REISERFS_DISK_OFFSET_IN_BYTES / size)) != size )
		{
			printk("reiserfs_read_super: unable to read superblock on dev %s\n",
				   kdevname(s->s_dev));
			device_close(fd);
			return 1;
		}
	
	rs = (struct reiserfs_super_block *)bh->b_data;

	if (strncmp (rs->s_magic,  REISERFS_SUPER_MAGIC_STRING, strlen ( REISERFS_SUPER_MAGIC_STRING)))
		{
			brelse (bh);
			printk ("reiserfs_read_super: can't find a reiserfs filesystem on dev %s.\n",
					kdevname(s->s_dev));
			device_close(fd);		
			return 1;
		}
	sb.s_blocksize = le16_to_cpu (rs->s_blocksize);
*/
	sb.s_blocksize_bits = 0;
	
	if (device_read(fd, (char *) &sb, sizeof(sb), sb_offset) !=  sizeof(sb))
		{
			printk("reiserfs sb read failed");
			device_close(fd);
			return (-1);
		}

	if( !strncmp (sb.s_magic,  REISERFS_SUPER_MAGIC_STRING, strlen ( REISERFS_SUPER_MAGIC_STRING)) )
		{
			if (!quiet)
				{
					printk("reiserfs_mount: bad magic 0x%x\n", sb.s_magic);
				}
			device_close(fd);
			return (-1);
		}
#ifdef DEBUG_REISERFS
	printk("Superblock\n");
	printk("   [%d] blocks\n", sb.s_block_count);
#endif

	/* Calculate direct/indirect block limits for this file system (blocksize
	 * dependent) */
	blocksize = REISERFS_BLOCK_SIZE(&sb);
	directlim = REISERFS_NDIR_BLOCKS - 1;
	ptrs_per_blk = blocksize / sizeof(unsigned int);
	ind1lim = ptrs_per_blk + directlim;
	ind2lim = (ptrs_per_blk * ptrs_per_blk) + directlim;





	SB_BUFFER_WITH_SB (&sb) = bh;
	SB_DISK_SUPER_BLOCK (&sb) = rs;
	sb.s_op = &reiserfs_sops;



	
	root_inode = iget ((char*) &sb, root_key.k_objectid);


	
#ifdef DEBUG_REISERFS
	printk("reiserfs_mount: returning successfully\n");
#endif

	return (0);
}

/* Read the specified inode from the disk and return it to the user.
 * Returns NULL if the inode can't be read...
 */
static struct reiserfs_inode *reiserfs_iget(int ino)
{
	int i;
	struct reiserfs_inode *ip;
	struct inode_table_entry *itp;
	int group;
	unsigned long offset;

#ifdef DEBUG_REISERFS
	printk("reiserfs_iget(): called for ino = %d\n", ino);
	printk("...inodes per group = %d\n", sb.s_inodes_per_group);
#endif

	ip = NULL;
	itp = NULL;
    for (i = 0; i < MAX_OPEN_FILES; i++)
		{
			if (inode_table[i].free)
				{
					itp = &(inode_table[i]);
					ip = &(itp->inode);
					break;
				}
		}
	if ((ip == NULL) || (itp == NULL))
		{
			printk("reiserfs_iget: no free inodes\n");
			return (NULL);
		}
	
	group = (ino - 1) / sb.s_inodes_per_group;
	offset = ((unsigned long) gds[group].bg_inode_table * blocksize)
	    +
	    (((ino - 1) % sb.s_inodes_per_group) *
	     sizeof(struct reiserfs_inode));

#ifdef DEBUG_REISERFS
	printk("...group = %d, gds[group].bg_inode_table = %d\n", group,
	       gds[group].bg_inode_table);
	printk("...reading inode at offset 0x%lx\n", offset);
#endif
	if (device_read(fd, (char *) ip, sizeof(struct reiserfs_inode), offset)
	    != sizeof(struct reiserfs_inode))
		{
			printk("reiserfs_iget: read error\n");
			return (NULL);
		}

	itp->free = 0;
	itp->inumber = ino;
	itp->old_mode = ip->i_mode;

#ifdef DEBUG_REISERFS
	printk("reiserfs_iget(): returning inode, size is %d\n", ip->i_size);
#endif

	return (ip);
}

/* Release our hold on an inode.  Since this is a read-only application,
 * don't worry about putting back any changes...
 */
static void reiserfs_iput(struct reiserfs_inode *ip)
{
	struct inode_table_entry *itp;

	/* Find and free the inode table slot we used... */
	itp = (struct inode_table_entry *) ip;

	itp->inumber = 0;
	itp->free = 1;
}


/* Map a block offset into a file into an absolute block number.
 * (traverse the indirect blocks if necessary).  Note: Double-indirect
 * blocks allow us to map over 64Mb on a 1k file system.  Therefore, for
 * our purposes, we will NOT bother with triple indirect blocks.
 *
 * The "allocate" argument is set if we want to *allocate* a block
 * and we don't already have one allocated.
 */
static int reiserfs_blkno(struct reiserfs_inode *ip, int blkoff, int allocate)
{
	unsigned int *lp;
	unsigned int *ilp;
	unsigned int *dlp;
	int blkno;
	unsigned long iblkno;
	unsigned long diblkno;


	ilp = (unsigned int *) iblkbuf;
	dlp = (unsigned int *) diblkbuf;

	if (allocate)
		{
			printk
				("reiserfs_blkno: Cannot allocate on a readonly file system!\n");
			return (0);
		}

	lp = (unsigned int *) blkbuf;

	/* If it's a direct block, it's easy! */
	if (blkoff <= directlim)
		{
			return (ip->i_block[blkoff]);
		}
	
	/* Is it a single-indirect? */
	if (blkoff <= ind1lim)
		{
			iblkno = ip->i_block[REISERFS_IND_BLOCK];
			
			if (iblkno == 0)
				{
					return (0);
				}
			
			/* Read the indirect block */
			if (cached_iblkno != iblkno)
				{
#ifdef DEBUG_REISERFS
					printk("Reading indirect block at 0x%lx\n",
						   (iblkno * blocksize));
#endif
					if (device_read
						(fd, iblkbuf, blocksize,
						 (iblkno * blocksize)) != blocksize)
						{
							printk("reiserfs_blkno: error on iblk read\n");
							return (0);
						}
					cached_iblkno = iblkno;
				}

			blkno = ilp[blkoff - (directlim + 1)];
			return (blkno);
		}

	/* Is it a double-indirect? */
	if (blkoff <= ind2lim)
		{

		/* Find the double-indirect block */
		diblkno = ip->i_block[REISERFS_DIND_BLOCK];
		
		if (diblkno == 0)
			{
				return (0);
			}
		
		/* Read in the double-indirect block */
		if (cached_diblkno != diblkno)
			{
#ifdef DEBUG_REISERFS
				printk("Reading double indirect block at 0x%lx\n",
					   (diblkno * blocksize));
#endif
				if (device_read
					(fd, diblkbuf, blocksize,
					 (diblkno * blocksize)) != blocksize)
					{
						printk
							("reiserfs_blkno: err reading dindr blk\n");
						return (0);
					}
				cached_diblkno = diblkno;
			}

		/* Find the single-indirect block pointer ... */
		iblkno = dlp[(blkoff - (ind1lim + 1)) / ptrs_per_blk];
		
		if (iblkno == 0)
			{
				return (0);
			}


		/* Read the indirect block */

		if (cached_iblkno != iblkno)
			{
				if (device_read
					(fd, iblkbuf, blocksize,
					 (iblkno * blocksize)) != blocksize)
					{
						printk("reiserfs_blkno: err on iblk read\n");
						return (0);
					}
				cached_iblkno = iblkno;
			}

		/* Find the block itself. */
		blkno = ilp[(blkoff - (ind1lim + 1)) % ptrs_per_blk];
		return (blkno);
		}
	
	if (blkoff > ind2lim)
		{
			printk("reiserfs_blkno: block number too large: %d\n", blkoff);
			return (0);
		}

	return (0);
}

/* Return the size of the open file */
static long reiserfs_fdsize(int fd)
{
	struct reiserfs_inode *ip;

	ip = &(inode_table[fd].inode);
	return ip->i_size;
}

/* Read block number "blkno" from the specified file */
static int reiserfs_bread(int fd, long blkno, char *buffer)
{
	struct reiserfs_inode *ip;

	ip = &(inode_table[fd].inode);
	reiserfs_breadi(ip, blkno, buffer);
	return 0;
}

static void reiserfs_breadi(struct reiserfs_inode *ip, long blkno, char *buffer)
{
	long dev_blkno;


#ifdef DEBUG_REISERFS
	printk("reiserfs_breadi(): called for block %ld\n", blkno);
#endif

	dev_blkno = reiserfs_blkno(ip, blkno, 0);
#ifdef DEBUG_REISERFS
	printk("reiserfs_breadi(): dev_blkno = %ld\n", dev_blkno);
#endif
	if (dev_blkno == 0)
		{

			/* This is a "hole" */
			memset(buffer, 0, blocksize);
		}
	else
		{
		
			/* Read it for real */
#ifdef DEBUG_REISERFS
			printk
				("reiserfs_breadi(): reading block %ld at offset 0x%lx\n",
				 blkno, (dev_blkno * blocksize));
#endif
			if (device_read
				(fd, buffer, blocksize,
				 (dev_blkno * blocksize)) != blocksize)
				{
					printk("reiserfs_bread: read error");
				}
		}
}


static struct reiserfs_inode *reiserfs_namei(char *name)
{
	char namebuf[256];
	char *component;
	struct reiserfs_inode *dir_inode = NULL, *file_inode;
	struct reiserfs_dir_entry *dp;
	int next_ino;
	int link_count = 0;


#ifdef DEBUG_REISERFS
	printk("reiserfs_namei(): called for %s\n", name);
#endif

	/* Squirrel away a copy of "namebuf" that we can molest */
	strcpy(namebuf, name);

	/* Start at the root... */
	/* We'll get back here to these nifty labels if we encounter a symlink */
      abs:
	file_inode = reiserfs_iget(REISERFS_ROOT_INO);
  rel:
	if (!file_inode)
		return NULL;

	component = strtok(namebuf, "/");
	while (component)
		{
			int diroffset;
			int blockoffset;
			int component_length;
			
		/* Search for the specified component in the current directory inode.
		 */

#ifdef MILO_SYMLINK_DIRS
			if (S_ISLNK(file_inode->i_mode))
				{
					if (++link_count > 10)
						printk("%s: possibly circular symlink\n",
				       name);
					else
						{
							char *t;
							
							/* Set blkbuf to symbolic link content */
							if (file_inode->i_blocks)
								reiserfs_breadi(file_inode, 0, blkbuf);
							else
								strcpy(blkbuf,
									   (char *) file_inode->
									   i_block);
							
							/* Add rest of the path to it */
							t = blkbuf + strlen(blkbuf);
							while (component != NULL)
								{
									t =
										stpcpy(stpcpy(t, "/"),
											   component);
									component = strtok(NULL, "/");
								}

				/* Put it back to namebuf */
							strcpy(namebuf, blkbuf);
#ifdef DEBUG_REISERFS
							printk
								("substituted symlink + rest of path: %s\n",
								 namebuf);
#endif

							if (*namebuf == '/')
								{
									/* Absolute symlink */
									if (dir_inode)
										reiserfs_iput(dir_inode);
									if (file_inode)
										reiserfs_iput(file_inode);
									goto abs;
								}
							else
								{
					/* Relative symlink */
									if (file_inode)
										reiserfs_iput(file_inode);
									file_inode = dir_inode;
									goto rel;
								}
						}
				}
#endif				/* MILO_SYMLINK_DIRS */

			if (dir_inode)
				reiserfs_iput(dir_inode);
			dir_inode = file_inode;
			file_inode = NULL;
			next_ino = -1;
			
			if (!S_ISDIR(dir_inode->i_mode))
				{
					printk("%s: path component is not a directory\n",
						   name);
					break;
				}

			component_length = strlen(component);
			diroffset = 0;
#ifdef DEBUG_REISERFS
			printk("...dir_inode->i_size = %d\n", dir_inode->i_size);
#endif
			while (diroffset < dir_inode->i_size)
				{
					blockoffset = 0;
#ifdef DEBUG_REISERFS
					printk("reiserfs_namei(): calling reiserfs_breadi\n");
#endif
					reiserfs_breadi(dir_inode, diroffset / blocksize,
									blkbuf);
					while (blockoffset < blocksize)
						{
							dp =
								(struct reiserfs_dir_entry *) (blkbuf +
															   blockoffset);
#if DEBUG_REISERFS
							{
								int i;
								
								printk("...name is ");
								for (i = 0; i < dp->name_len; i++)
									printk("%c", dp->name[i]);
								printk("\n");
							}
#endif
							if ((dp->name_len == component_length) &&
								(strncmp(component, dp->name,
										 component_length) == 0))
								{

									/* Found it! */
									next_ino = dp->inode;
									break;
								}
							
							/* Go to next entry in this block */
							blockoffset += dp->rec_len;
						}
					if (next_ino >= 0)
						{
							break;
						}

					/* If we got here, then we didn't find the component. Try the next 
					 * block in this directory... */
					diroffset += blocksize;
				}		/* end-while */

			/* If next_ino is negative, then we've failed (gone all the way through
			 * without finding anything) */
			if (next_ino < 0)
				{
#ifdef DEBUG_REISERFS
					printk("...we failed\n");
#endif
					break;
				}

			/* Otherwise, we can get this inode and find the next component
			 * string... */
			file_inode = reiserfs_iget(next_ino);
			
			component = strtok(NULL, "/");
		}

	/* If we get here, then we got through all the components. Whatever we got 
	 * must match up with the last one. */
	if (file_inode && S_ISLNK(file_inode->i_mode)) 	/* Is it a symlink? */
		{
			if (++link_count > 5)
				printk("%s: possibly circular symlink\n", name);
			else
				{
					if (file_inode->i_blocks)
						{
							reiserfs_breadi(dir_inode, 0, blkbuf);
							strncpy(namebuf, blkbuf,
									sizeof namebuf - 1);
						}
					else
						strncpy(namebuf,
								(char *) file_inode->i_block,
								sizeof namebuf - 1);
					reiserfs_iput(file_inode);
					if (*namebuf == '/') 	/* absolute */
						{
							reiserfs_iput(dir_inode);
							goto abs;
						}	/* otherwise it's relative */
					file_inode = dir_inode;
					goto rel;
				}
		}
	if (dir_inode)
		reiserfs_iput(dir_inode);
	return (file_inode);
}

static void reiserfs_ls(char *dirname, char *devname)
{
	char filename[512];
	char format_string[16];
	struct reiserfs_inode *dir_inode;
	struct reiserfs_inode *file_inode;
	struct reiserfs_dir_entry *dp;
	int diroffset;
	int blockoffset;
	int column_size = 0;
	int columns;
	int col;
	char typetag;


#ifdef DEBUG_REISERFS
	printk("reiserfs_ls(): called for %s\n", dirname);
#endif

	/* Get the inode for the specified directory */
	dir_inode = reiserfs_namei(dirname);
	if (!dir_inode)
		{
			printk("%s: no such file or directory\n", dirname);
			return;
		}

	/* If this inode exists but is *not* a directory, just list it here */
	if ((dir_inode->i_mode & S_IFMT) != S_IFDIR)
		{
#ifdef BOOT_VERBOSE
			printk("Directory of: %s:/%s\n", devname, dirname);
#endif
			printk("%s\n", dirname);
			reiserfs_iput(dir_inode);
			return;
		}

	/* First pass: determine maximum column size */

	diroffset = 0;
	while (diroffset < dir_inode->i_size)
		{
			blockoffset = 0;
			reiserfs_breadi(dir_inode, diroffset / blocksize, blkbuf);
			while (blockoffset < blocksize)
				{
					dp =
						(struct reiserfs_dir_entry *) (blkbuf +
													   blockoffset);
					if (dp->name_len > column_size)
						{
							column_size = dp->name_len;
						}
					blockoffset += dp->rec_len;
				}

		/* Go to the next  block in this directory... */
			diroffset += blocksize;
		}			/* end-while */
	
	/* Compute the format string for this listing... */
	column_size += 3;	/* Add spacer and type tag */
	if (column_size >= 79)
		{
			columns = 1;
			strcpy(format_string, "%s");
		}
	else
		{
			columns = 79 / column_size;
			column_size = 79 / columns;
			sprintf(format_string, "%%-%ds", column_size);
		}

	/* Second pass: Get all the filenames and modes and
	 * print them out...
	 */
	diroffset = 0;
	col = 0;
#ifdef BOOT_VERBOSE
	printk("Directory of: %s:/%s\n", devname, dirname);
#endif
	while (diroffset < dir_inode->i_size)
		{
			blockoffset = 0;
			reiserfs_breadi(dir_inode, diroffset / blocksize, blkbuf);
			while (blockoffset < blocksize)
				{
					dp =
						(struct reiserfs_dir_entry *) (blkbuf +
													   blockoffset);
					if (dp->name_len > 0)
						{
							strncpy(filename, dp->name, dp->name_len);
							file_inode = reiserfs_iget(dp->inode);
							if (file_inode)
								{
									switch (file_inode->
											i_mode & S_IFMT)
										{
										case S_IFLNK:
											typetag = '@';
											break;
										case S_IFDIR:
											typetag = '/';
											break;
										case S_IFIFO:
										case S_IFSOCK:
											typetag = '=';
											break;
										default:
											typetag = ' ';
											break;
										}
									reiserfs_iput(file_inode);
								}
							else
								{
									typetag = ' ';
								}
							filename[dp->name_len] = typetag;
							filename[dp->name_len + 1] = '\0';
							printk(format_string, filename);
							if ((++col % columns) == 0) {
								printk("\n");
							}
						}
					blockoffset += dp->rec_len;
				}
			
			/* Go to the next  block in this directory... */
			diroffset += blocksize;
		}			/* end-while */
	
	if ((col % columns) != 0)
		{
			printk("\n");
		}

	/* We're done with the directory now... */
	reiserfs_iput(dir_inode);

}

static int reiserfs_open(char *filename)
{

	/* Unix-like open routine.  Returns a small integer (actually an index into
	 * the inode table... */
	struct reiserfs_inode *ip;

#ifdef DEBUG_REISERFS
	printk("reiserfs_open() called with filename = %s\n", filename);
#endif
	ip = reiserfs_namei(filename);
#ifdef DEBUG_REISERFS
	printk("...namei returned 0x%p\n", ip);
#endif

	if (ip)
		{
			struct inode_table_entry *itp;
			
			itp = (struct inode_table_entry *) ip;
			return (itp - inode_table);
		}
	else
		{
			return (-1);
		}
}


static void reiserfs_close(int fd)
{
	reiserfs_iput(&(inode_table[fd].inode));
#ifdef DEBUG_REISERFS
	printk("reiserfs_Closing device, fd=%d\n", fd);
#endif
	device_close(fd);
}


        static char p[] = "abcdefghijklmnopkrstuvwxyz01234567890";
		static char p2[40];
		int j;
		for(i=0;i<8;i++)
			{
				printk ("\nDEBUG: memcpy %d test: %s\n", i,p2);
				for(j=0;j<8;j++)
					{
//						memset(p2, '.',40 );
						memcpy(p + 8+i, p + j, 17);
						printk ("DEBUG: memcpy %d test: %s\n", j,p2);
					}
				kbd_getc();	
			}





	
